/*
  LCD File Manager

  T. Nakagawa
  2014/01/18
*/


#include <dirent.h>
#include <locale.h>
#include <ncursesw/ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>


#define BUF_SIZE 1024
#define EXT_SIZE 256


/*
  Shorten the given string in UTF-8 to fit in the specified width.
*/
void clip(char *str, int size) {
  int i;
  int len;

  len = 0;
  for (i = 0; i < strlen(str); ) {
    int width;
    if ((str[i] & 0x80) == 0) {
      width = 1;
    } else {
      width = 2;
    }
    if (len + width > size) break;
    len += width;
    if ((str[i] & 0x80) == 0) {
      i++;
    } else {
      i++;
      while (i < strlen(str) && (str[i] & 0x80) != 0 && (str[i] & 0x40) == 0) i++;
    }
  }
  str[i] = '\0';
  return;
}


void basename(char *dst, char *src) {
  char *ptr;
  strcpy(dst, src);
  ptr = strrchr(dst, '.');
  if (ptr != NULL) {
    *ptr = '\0';
  }
  return;
}


void extname(char *dst, char *src) {
  char *ptr;
  ptr = strchr(src, '.');
  if (ptr == NULL) {
    strcpy(dst, "");
  } else {
    strcpy(dst, ptr + 1);
  }
  return;
}


int getlist(char *path, struct dirent ***list, int *num) {
  int i, j;
  struct dirent **tmplist;
  int tmpnum;

  tmpnum = scandir(path, &tmplist, NULL, alphasort);
  if (tmpnum >= 0) {
    /* Remove dot files. */
    j = 0;
    for (i = 0; i < tmpnum; i++) {
      if (tmplist[i]->d_name[0] == '.') {
	free(tmplist[i]);
      } else {
	tmplist[j++] = tmplist[i];
      }
    }
    tmpnum = j;

    if (*list != NULL) {
      for (i = 0; i < *num; i++) {
	free((*list)[i]);
      }
      free(*list);
    }
    *list = tmplist;
    *num = tmpnum;
    return 0;
  } else {
    return 1;
  }
}


void open_curses(WINDOW **win) {
  setlocale(LC_ALL, "");
  initscr();
  start_color();
  curs_set(0);
  noecho();
  init_pair(1, COLOR_MAGENTA, COLOR_YELLOW);
  init_pair(2, COLOR_WHITE, COLOR_BLACK);
  init_pair(3, COLOR_CYAN, COLOR_BLACK);
  init_pair(4, COLOR_WHITE, COLOR_BLUE);
  init_pair(5, COLOR_CYAN, COLOR_BLUE);
  init_pair(6, COLOR_RED, COLOR_GREEN);
  win[0] = newwin(LINES, COLS, 0, 0);
  win[1] = subwin(win[0], 1, COLS, 0, 0);
  wbkgd(win[1], COLOR_PAIR(1));
  win[2] = subwin(win[0], LINES - 2, COLS, 1, 0);
  wbkgd(win[2], COLOR_PAIR(2));
  win[3] = subwin(win[0], 1, COLS - 1, LINES - 1, 0);
  wbkgd(win[3], COLOR_PAIR(6));
  refresh();
}


void close_curses(void) {
  endwin();
}


int main(int argc, char **argv) {
  int i;
  char *conffile;
  char *root;
  char path[1024];
  int num;
  struct dirent **list;
  int ext_num;
  char *ext[EXT_SIZE][2];
  char buf[BUF_SIZE];
  int dirty;
  FILE *fp;
  int cursor;
  int offset;
  WINDOW *win[4];
  char padding[BUF_SIZE];
  int hist[BUF_SIZE / 2];	/* Cursor position history. */
  int hist_num;

  if (argc != 3) {
    fprintf(stderr, "usage: %s <configuration file> <root directory>\n", argv[0]);
    return 1;
  }
  conffile = argv[1];
  root = argv[2];

  /* Read a configuration file */
  if ((fp = fopen(conffile, "r")) == NULL) {
    fprintf(stderr, "ERROR: Cannot read the configuration file.\n");
    return 1;
  }
  ext_num = 0;
  while (fgets(buf, BUF_SIZE, fp) != NULL) {
    char *ptr;
    if (ext_num >= EXT_SIZE) {
      fprintf(stderr, "ERROR: Too many extensions.\n");
      return 1;
    }

    buf[strlen(buf) - 1] = '\0';
    ptr = strchr(buf, '\t');
    if (ptr == NULL) {
      fprintf(stderr, "ERROR: Wrong configuration file.\n");
      return 1;
    }
    *ptr++ = '\0';
    ext[ext_num][0] = strdup(buf);
    ext[ext_num][1] = strdup(ptr);
    ext_num++;
  }
  fclose(fp);

  strcpy(path, root);
  list = NULL;
  if (getlist(path, &list, &num)) {
    fprintf(stderr, "ERROR: Cannot list the directory %s\n", path);
    return 1;
  }

  open_curses(win);
  for (i = 0; i < COLS; i++) padding[i] = ' ';
  padding[COLS] = '\0';
  cursor = 0;
  offset = 0;
  dirty = 1;

  hist_num = 0;
  for (; ; ) {
    char c;

    if (dirty) {
      /* Title window */
      strcpy(buf, path + strlen(root));
      clip(buf, COLS);
      mvwaddstr(win[1], 0, 0, buf);
      wclrtoeol(win[1]);
      wrefresh(win[1]);

      /* Status window */
      if (list[cursor]->d_type == DT_DIR) {
	strcpy(buf, "");
      } else {
	char *ptr;
	ptr = strrchr(list[cursor]->d_name, '.');
	if (ptr == NULL) {
	  strcpy(buf, "");
	} else {
	  strcpy(buf, ptr + 1);
	}
      }
      clip(buf, COLS - 1);
      mvwaddstr(win[3], 0, 0, buf);
      wclrtoeol(win[3]);
      sprintf(buf, "%d/%d", cursor + 1, num);
      buf[COLS - 1] = '\0';
      mvwaddstr(win[3], 0, COLS - strlen(buf) - 1, buf);
      wrefresh(win[3]);

      /* Main window */
      for (i = 0; i < COLS - 2; i++) {
	wmove(win[2], i, 0);
	if (offset + i >= num) {
	  wclrtobot(win[2]);
	  break;
	}
	strcpy(buf, list[offset + i]->d_name);
	if (list[offset + i]->d_type == DT_DIR) {
	  if (i == cursor - offset) {
	    wattron(win[2], COLOR_PAIR(5));
	  } else {
	    wattron(win[2], COLOR_PAIR(3));
	  }
	} else {
	  basename(buf, buf);
	  if (i == cursor - offset) {
	    wattron(win[2], COLOR_PAIR(4));
	  } else {
	    wattron(win[2], COLOR_PAIR(2));
	  }
	}
	strcat(buf, padding);
	clip(buf, COLS);
	waddstr(win[2], buf);
	wclrtoeol(win[2]);
      }
      wrefresh(win[2]);
      refresh();
      dirty = 0;
    }

    c = getch();
    if (c == 'q') {
      break;
    } else if (c == 'h' && strlen(path) > strlen(root)) {	/* Left */
      /* Parent directory */
      strcpy(buf, path);
      *strrchr(buf, '/') = '\0';
      if (!getlist(buf, &list, &num)) {
	strcpy(path, buf);
	cursor = hist[--hist_num];
	if (cursor >= num) cursor = num - 1;
	offset = 0;
	while (cursor - offset >= LINES - 2) offset += LINES - 2;
	dirty = 1;
      }
    } else if (c == 'l') {	/* Right */
      /* Child directory or execution */
      if (list[cursor]->d_type == DT_DIR) {
	strcpy(buf, path);
	strcat(buf, "/");
	strcat(buf, list[cursor]->d_name);
	if (!getlist(buf, &list, &num)) {
	  strcpy(path, buf);
	  hist[hist_num++] = cursor;
	  cursor = 0;
	  offset = 0;
	  dirty = 1;
	}
      } else {
	extname(buf, list[cursor]->d_name);
	for (i = 0; i < ext_num; i++) {
	  if (strcmp(buf, ext[i][0]) == 0) {
	    int pid;

	    close_curses();
	    pid = fork();
	    if (pid == 0) {	/* Child process. */
	      strcpy(buf, path);
	      strcat(buf, "/");
	      strcat(buf, list[cursor]->d_name);
	      execl(ext[i][1], ext[i][1], buf, NULL);
	      return 1;
	    } else if (pid > 0) {	/* Parent process. */
	      wait(NULL);
	    }
	    open_curses(win);
	    dirty = 1;
	    break;
	  }
	}
      }
    } else if (c == 'j' && cursor + 1 < num) {
      /* Down */
      cursor++;
      if (cursor - offset >= LINES - 2) {
	offset += LINES - 2;
      }
      dirty = 1;
    } else if (c == 'k' && cursor - 1 >= 0) {
      /* Up */
      cursor--;
      if (cursor - offset < 0) {
	offset -= LINES - 2;
      }
      dirty = 1;
    }
  }

  close_curses();

  return 0;
}
